package it.unibo.ai.rec.ui;
import it.unibo.ai.rec.common.TimeGranularity;
import it.unibo.ai.rec.model.ClosedMVI;
import it.unibo.ai.rec.model.Fluent;
import it.unibo.ai.rec.model.FluentGroup;
import it.unibo.ai.rec.model.FluentState;
import it.unibo.ai.rec.model.FluentsModel;
import it.unibo.ai.rec.model.HappenedEventSet;
import it.unibo.ai.rec.model.MVI;
import it.unibo.ai.rec.model.OpenMVI;
import it.unibo.ai.rec.model.RecTrace;
import it.unibo.ai.rec.model.StandardFluent;
import it.unibo.ai.rec.model.TrendMetric;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Paint;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.io.FileInputStream;
import java.io.InputStream;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.SwingUtilities;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.annotations.XYTextAnnotation;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.labels.ItemLabelAnchor;
import org.jfree.chart.labels.ItemLabelPosition;
import org.jfree.chart.labels.StandardCategoryItemLabelGenerator;
import org.jfree.chart.labels.StandardXYItemLabelGenerator;
import org.jfree.chart.labels.XYItemLabelGenerator;
import org.jfree.chart.plot.CombinedDomainXYPlot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.SeriesRenderingOrder;
import org.jfree.chart.plot.ValueMarker;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYAreaRenderer;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.ui.RectangleInsets;
import org.jfree.ui.TextAnchor;

public class FluentChartFactory {
	
	
	
	public static Font labelFont = new Font("Lucida Grande",Font.BOLD,14);
	public static Font titleFont = new Font("Lucida Grande", Font.BOLD, 14);
	public static Font annotationFont = new Font("Lucida Grande", Font.BOLD, 13);
	public static Font axisFont = new Font("Lucida Grande", Font.BOLD, 14);

	
	
	private static int START_TIME = -1;
	private static String START_NAME = "start";
	
	
	private static double BASE = 0;
	private static double STEP = 1;
	private static double DELTA = 0.00000001;
	
	private static int INSET = -3;

	
	private static Paint 	DEFAULT_GLOBAL_BG_PAINT = Color.white;
	private static Paint 	DEFAULT_CHART_BG_PAINT 	= Color.white;
	
	private static Stroke 	DEFAULT_MARKER_STROKE 	= new BasicStroke(1);
	private static Paint 	DEFAULT_MARKER_PAINT 	= Color.black;
	
	private static Paint 	DEFAULT_GRIDLINE_PAINT 	= Color.black;
	private static float 	DEFAULT_ALPHA 			= 0.3f;
	
	private static Font		DEFAULT_FLUENT_FONT		= new Font("", Font.PLAIN, 10);
	

	private static int		DEFAULT_UNIT_WIDTH		= 35;
	private static int		DEFAULT_UNIT_HEIGHT		= 30;
	
	
	/*public static enum ChartModality {
		SUMMARY,
		INSTANCES,
		BOTH;
		
		public boolean showInstances() {
			return this.equals(INSTANCES) || this.equals(BOTH);
		}
		
		public boolean showSummary() {
			return this.equals(SUMMARY) || this.equals(BOTH);
		}
	}*/
	
	private ColorMap stateMap;
	private ColorMap fluentMap;
	
	private Paint globalBackgroundPaint;
	private Paint chartBackgroundPaint;
	private Paint gridlinePaint;

	private Stroke markerStroke;
	private Paint markerPaint;
	
	private float alpha;
	
	//private Font fluentFont;
	
	private int unitWidth;
	private int unitHeight;
	private int chartsGap;
	
	
	private TimeGranularity granularity;
	private EventOutputter eventOutputter;
	private boolean metricDistances;
	
	private static ColorMap createMap(String location) throws Exception {
		ColorMap map = new ColorMap();
		//map.load(FluentChartFactory.class.getResourceAsStream(location));
		InputStream in = new FileInputStream(new java.io.File(location).getAbsolutePath());
		map.load(in);
		in.close();
		return map;
	}
	
	public FluentChartFactory(TimeGranularity granularity, EventOutputter eventOutputter,  boolean metricDistances) throws Exception {
		this(
			granularity,
			eventOutputter,
			metricDistances, 
			createMap("/it/unibo/ai/rec/visualization/stateColors.properties"),
			createMap("/it/unibo/ai/rec/visualization/fluentColors.properties")
		);
	}
	
	public FluentChartFactory(TimeGranularity granularity, EventOutputter eventOutputter,  boolean metricDistances, String stateMapLocation, String fluentMapLocation) throws Exception {
		this(
			granularity,
			eventOutputter,
			metricDistances, 
			createMap(stateMapLocation),
			createMap(fluentMapLocation)
		);
	}
	
	public FluentChartFactory(TimeGranularity granularity, EventOutputter eventOutputter, boolean metricDistances, ColorMap stateMap, ColorMap fluentMap) throws Exception {
		
		this.stateMap = stateMap;
		this.fluentMap = fluentMap;
		
		globalBackgroundPaint = DEFAULT_GLOBAL_BG_PAINT;
		chartBackgroundPaint = DEFAULT_CHART_BG_PAINT;
		gridlinePaint = DEFAULT_GRIDLINE_PAINT;

		markerStroke = DEFAULT_MARKER_STROKE;
		markerPaint = DEFAULT_MARKER_PAINT;
		
		alpha = DEFAULT_ALPHA;
		
		//fluentFont = DEFAULT_FLUENT_FONT;
		
		unitWidth = DEFAULT_UNIT_WIDTH;
		unitHeight = DEFAULT_UNIT_HEIGHT;
		
		this.granularity = granularity;
		this.eventOutputter = eventOutputter;
		this.metricDistances = metricDistances;
	}
	
	
	//public Font getFluentFont() {
	//	return fluentFont;
	//}

	//public void setFluentFont(Font fluentFont) {
	//	this.fluentFont = fluentFont;
	//}

	public Paint getGlobalBackgroundPaint() {
		return globalBackgroundPaint;
	}

	public void setGlobalBackgroundPaint(Paint globalBackgroundPaint) {
		this.globalBackgroundPaint = globalBackgroundPaint;
	}

	public Paint getChartBackgroundPaint() {
		return chartBackgroundPaint;
	}

	public void setChartBackgroundPaint(Paint chartBackgroundPaint) {
		this.chartBackgroundPaint = chartBackgroundPaint;
	}

	public Paint getGridlinePaint() {
		return gridlinePaint;
	}

	public void setGridlinePaint(Paint gridlinePaint) {
		this.gridlinePaint = gridlinePaint;
	}

	public Stroke getMarkerStroke() {
		return markerStroke;
	}

	public void setMarkerStroke(Stroke markerStroke) {
		this.markerStroke = markerStroke;
	}

	public Paint getMarkerPaint() {
		return markerPaint;
	}

	public void setMarkerPaint(Paint markerPaint) {
		this.markerPaint = markerPaint;
	}
	
	public float getAlpha() {
		return alpha;
	}

	public void setAlpha(float alpha) {
		this.alpha = alpha;
	}
	
	public TimeGranularity getGranularity() {
		return granularity;
	}

	public void setGranularity(TimeGranularity granularity) {
		this.granularity = granularity;
	}

	public EventOutputter getEventOutputter() {
		return eventOutputter;
	}

	public void setEventOutputter(EventOutputter eventOutputter) {
		this.eventOutputter = eventOutputter;
	}

	public boolean isMetricDistances() {
		return metricDistances;
	}

	public void setMetricDistances(boolean metricDistances) {
		this.metricDistances = metricDistances;
	}


	public int getWidth(RecTrace trace, ValueAxis axis) {
		int width = 0;
		double first = getX(trace.getHappenedEvents().get(0).getTime(), 1);
		double last = getX(
				trace.getHappenedEvents().lastElement().getTime(),
				trace.getHappenedEvents().size());
		if(metricDistances)	 {
			double minDiff = getMinDiff(trace, granularity, metricDistances);
			double maxDiff = last-first;
			width += (int)(maxDiff/minDiff)*unitWidth;
		}
		else
			width += (trace.getHappenedEvents().size())*unitWidth;
		return width;
	}
	
	
	public ChartPanel getSummaryFluentChart(RecTrace trace, FluentGroup group, boolean inverted) {
		return getFluentChart(trace, group, true, inverted);
	}
	
	public ChartPanel getDetailedFluentChart(RecTrace trace, FluentGroup group, boolean inverted) {
		return getFluentChart(trace, group, false, inverted);
	}

	public ChartPanel getFluentChart(RecTrace extTrace, FluentGroup group, boolean isSummary, boolean inverted) {
		RecTrace trace = new RecTrace(true);
		trace.getHappenedEvents().addAll(extTrace.getHappenedEvents());
		trace.addHappenedEvent(START_NAME,START_TIME);
		//double first = getX(trace.getEvents().get(0).getTime(), 0);//MARCO 1
		double first = START_TIME;
		double last = getX(
				trace.getHappenedEvents().lastElement().getTime(),
				trace.getHappenedEvents().size()-1);
		XYPlot fluentChartPlot = null;
		XYItemRenderer fluentChartRenderer = null;
		XYSeriesCollection seriesCollection = null;
		
		JFreeChart fluentChart = ChartFactory.createXYAreaChart(
												null, // chart title 
												null, // x axis label 
												//false,
												"", // y axis label 
												null, // data 
												PlotOrientation.VERTICAL, 
												false, // include legend 
												true, // tooltips 
												false // urls 
												);
		ChartPanel fluentChartPanel = new ChartPanel(fluentChart);
		fluentChartPlot = fluentChart.getXYPlot();
		ValueAxis domainAxis = createTraceAxis(trace);
		fluentChartPlot.setDomainAxis(domainAxis);
		fluentChartRenderer = fluentChartPlot.getRenderer();
		fluentChartPlot.setBackgroundPaint(this.getChartBackgroundPaint());
		fluentChartPlot.setDomainGridlinePaint(this.getGridlinePaint());
		fluentChartPlot.setRangeGridlinesVisible(false);
		if(isSummary) {
			fluentChartPlot.setForegroundAlpha(alpha);
			fluentChartPlot.setSeriesRenderingOrder(SeriesRenderingOrder.FORWARD);
		}
		else
			fluentChartPlot.setForegroundAlpha(1);
		
		seriesCollection = new XYSeriesCollection();
		fluentChartPlot.setDataset(seriesCollection);
		
		
		ValueAxis rangeAxis = 	fluentChartPlot.getRangeAxis();
		rangeAxis.setLowerBound(0);
		if(isSummary)
			rangeAxis.setUpperBound(STEP);
		else
			rangeAxis.setUpperBound(BASE+group.getFluents().size()*STEP);
		
		//rangeAxis.setLabelFont(getFluentFont());
		rangeAxis.setAxisLineVisible(false);
		rangeAxis.setTickLabelsVisible(false);
		rangeAxis.setTickMarksVisible(false);
		rangeAxis.setLabel(group.getName());
		rangeAxis.setLabelAngle(Math.PI/2);
		rangeAxis.setInverted(inverted);
	
		//rangeAxis.setLabelFont(labelFont);
		
		double low = BASE;
		double high = low + STEP;
		int seriesNo = 0;
		
		List<Fluent> fluents = new ArrayList<Fluent>(group.getFluents().values());
		Collections.sort(fluents);
		
		if(isSummary) {
			String s = "";
			if(fluents.size() > 1)
				s = "s";
			XYTextAnnotation hintAnnotation = new XYTextAnnotation(fluents.size()+" instance"+s, 0, 0);
			hintAnnotation.setTextAnchor(TextAnchor.TOP_LEFT);
			hintAnnotation.setFont(annotationFont);
			fluentChartPlot.addAnnotation(hintAnnotation);
		}
		
		
		
		int position = 0;
		for(Fluent fluent: fluents) {
			fluentChartPlot.addRangeMarker(new ValueMarker(low, markerPaint, markerStroke));
			
			position++;
			if(!isSummary) {
				XYTextAnnotation instanceAnnotation = new XYTextAnnotation(""+position, 0, low);
				instanceAnnotation.setTextAnchor(TextAnchor.TOP_LEFT);
				instanceAnnotation.setFont(annotationFont);
				fluentChartPlot.addAnnotation(instanceAnnotation);
			}		
			
			
			Map<String,FluentState> states = fluent.getStates();
			Set<String> stateNames = states.keySet();
			for(String stateName: stateNames) {
				Color seriesPaint = null;
				if(fluent instanceof StandardFluent) {
					seriesPaint = fluentMap.getColor(fluent.getName());
					if(seriesPaint == null)
						seriesPaint = fluentMap.getColor("default");
				}
				else {
					seriesPaint = stateMap.getColor(stateName);
					if(seriesPaint == null)
						seriesPaint = stateMap.getColor("default");					
				}
				fluentChartRenderer.setSeriesPaint(seriesNo, seriesPaint);
				fluentChartRenderer.setSeriesStroke(seriesNo, new BasicStroke(3));
				XYSeries stateSeries = new XYSeries(stateName, true, false);
				FluentState state = states.get(stateName);
				MVI[] mvis = state.getMVIs();
				for(int j = 0; j < mvis.length; j++) {
					MVI mvi = mvis[j];
					double xStart = getX(mvi.getStart(), getPosition(trace, mvi.getStart()));
					if(!isSummary) {
						XYTextAnnotation stateAnnotation = new XYTextAnnotation(stateName, xStart, high);
						stateAnnotation.setTextAnchor(inverted?TextAnchor.BOTTOM_LEFT:TextAnchor.TOP_LEFT);
						stateAnnotation.setFont(annotationFont);
						fluentChartPlot.addAnnotation(stateAnnotation);
					}
					stateSeries.addOrUpdate(xStart-DELTA,low);
					stateSeries.addOrUpdate(xStart+DELTA,high);
					
					double xEnd = 0;
					
					if(mvi instanceof OpenMVI)
						xEnd = last+1-DELTA;//+1 ???
					else if(mvi instanceof ClosedMVI) {
						ClosedMVI closedMVI = (ClosedMVI)mvi;
						xEnd = getX(closedMVI.getEnd(), getPosition(trace, closedMVI.getEnd()));
					}
					stateSeries.addOrUpdate(xEnd-DELTA, high);
					stateSeries.addOrUpdate(xEnd+DELTA,low);
			
				}
				seriesCollection.addSeries(stateSeries);

				seriesNo++;
			}
			if(!isSummary) {
				low = high;
				high = low + STEP;
			}
		}
		
		int height = 0;//((HeightEvaluator)domainAxis).getHeight(fluentChartPanel.getFontMetrics(domainAxis.getLabelFont()));
		
		int factor = 1;
		if(!isSummary)
			factor = group.getFluents().size();
		height += factor*unitHeight;

		
		//int width = SwingUtilities.computeStringWidth(fluentChartPanel.getFontMetrics(labelFont), longestLabel);
		
		
		setDimensions(fluentChartPanel, getWidth(trace, domainAxis), height);
		return fluentChartPanel;
	}
	
	
	public ChartPanel getMetricChart(RecTrace extTrace, FluentsModel monitoringState, TrendMetric metric, int height) {
		RecTrace trace = new RecTrace(true);
		trace.getHappenedEvents().addAll(extTrace.getHappenedEvents());
		trace.addHappenedEvent(START_NAME,START_TIME);
		JFreeChart metricChart = ChartFactory.createXYLineChart(
				metric.getName(), // chart title 
				null, // x axis label 
				//false,
				"", // y axis label 
				null, // data 
				PlotOrientation.VERTICAL, 
				false, // include legend 
				true, // tooltips 
				true // urls 
				);
		metricChart.getTitle().setFont(titleFont);
		ChartPanel metricChartPanel = new ChartPanel(metricChart);
		XYPlot metricChartPlot = metricChart.getXYPlot();
		ValueAxis domainAxis = createTraceAxis(trace);
		metricChartPlot.setDomainAxis(domainAxis);
		XYItemRenderer metricChartRenderer = metricChartPlot.getRenderer();
		metricChartPlot.setBackgroundPaint(this.getChartBackgroundPaint());
		metricChartPlot.setDomainGridlinePaint(this.getGridlinePaint());
		metricChartPlot.setRangeGridlinesVisible(false);
		metricChartPlot.getRangeAxis().setRange(0, 1);
		metricChartPlot.getRangeAxis().setVisible(false);
		
		XYSeries stateSeries = new XYSeries(metric.getName(), true, false);
		int i = 0;
		double x = 0;
		double y = 0;
		for(HappenedEventSet e: trace.getHappenedEvents()) {
			x = getX(e.getTime(), i);
			y = metric.getValue(monitoringState, e.getTime());
			stateSeries.add(x, y);
			i++;
		}
		stateSeries.add(getX(trace.getHappenedEvents().lastElement().getTime()+1, i+1),y);
		XYSeriesCollection series = new XYSeriesCollection();
		series.addSeries(stateSeries);
		metricChartPlot.setDataset(series);
	
		DecimalFormat metricValueFormat = new DecimalFormat("#.##");
		DateFormat df = null;
		metricChartRenderer.setBaseItemLabelGenerator(new StandardXYItemLabelGenerator("{2}", metricValueFormat, metricValueFormat));  
		metricChartRenderer.setBaseItemLabelsVisible(true);
		metricChartRenderer.setBasePositiveItemLabelPosition(
			      new ItemLabelPosition(ItemLabelAnchor.CENTER, TextAnchor.TOP_CENTER));
		metricChartRenderer.setBaseItemLabelFont(annotationFont);
		setDimensions(metricChartPanel, getWidth(trace, domainAxis), height);
		
		return metricChartPanel;
	}
	
	/*
	public ChartPanel getTimingsChart(RecTrace extTrace, long[] timings,  int height) {
		RecTrace trace = new RecTrace(true);
		trace.getHappenedEvents().addAll(extTrace.getHappenedEvents());
		trace.addHappenedEvent(START_NAME,START_TIME);
		JFreeChart timingsChart = ChartFactory.createBarChart(
				"", // chart title 
				null, // x axis label 
				//false,
				"", // y axis label 
				null, // data 
				PlotOrientation.VERTICAL, 
				false, // include legend 
				true, // tooltips 
				true // urls 
				);
		//metricChart.getTitle().setFont(titleFont);
		ChartPanel timingsChartPanel = new ChartPanel(timingsChart);
		XYPlot timingsChartPlot = timingsChart.getXYPlot();
		ValueAxis domainAxis = createTraceAxis(trace);
		timingsChartPlot.setDomainAxis(domainAxis);
		XYItemRenderer metricChartRenderer = timingsChartPlot.getRenderer();
		timingsChartPlot.setBackgroundPaint(this.getChartBackgroundPaint());
		timingsChartPlot.setDomainGridlinePaint(this.getGridlinePaint());
		timingsChartPlot.setRangeGridlinesVisible(false);
		//timingsChartPlot.getRangeAxis().setRange(0, 1);
		timingsChartPlot.getRangeAxis().setVisible(false);
		
		XYSeries stateSeries = new XYSeries("", true, false);
		int i = 0;
		double x = 0;
		double y = 0;
		for(HappenedEventSet e: trace.getHappenedEvents()) {
			x = getX(e.getTime(), i);
			y = timings[i];
			stateSeries.add(x, y);
			i++;
		}
		stateSeries.add(getX(trace.getHappenedEvents().lastElement().getTime()+1, i+1),y);
		XYSeriesCollection series = new XYSeriesCollection();
		series.addSeries(stateSeries);
		timingsChartPlot.setDataset(series);
	
		//DecimalFormat metricValueFormat = new DecimalFormat("#.##");
		//DateFormat df = null;
		metricChartRenderer.setBaseItemLabelGenerator(new StandardXYItemLabelGenerator("{2}", metricValueFormat, metricValueFormat));  
		metricChartRenderer.setBaseItemLabelsVisible(true);
		metricChartRenderer.setBasePositiveItemLabelPosition(
			      new ItemLabelPosition(ItemLabelAnchor.CENTER, TextAnchor.TOP_CENTER));
		metricChartRenderer.setBaseItemLabelFont(annotationFont);
		setDimensions(metricChartPanel, getWidth(trace, domainAxis), height);
		
		return metricChartPanel;
	}
	*/
	
	
	public void setDimensions(ChartPanel panel, int width, int height) {
		panel.getChart().getPlot().setInsets(new RectangleInsets(INSET,INSET,INSET,INSET));
		Dimension d = new Dimension(width,height);
		panel.setMaximumDrawWidth(width);
		panel.setMinimumDrawWidth(width);
		panel.setMaximumDrawHeight(height);
		panel.setMinimumDrawHeight(height);
		panel.setPreferredSize(d);
		panel.setMaximumSize(d);
		panel.setMinimumSize(d);
	}
	

	
	public int getUnitWidth() {
		return unitWidth;
	}


	public void setUnitWidth(int unitWidth) {
		this.unitWidth = unitWidth;
	}


	public int getUnitHeight() {
		return unitHeight;
	}


	public void setUnitHeight(int unitHeight) {
		this.unitHeight = unitHeight;
	}


	public ValueAxis createTraceAxis(RecTrace trace) {
		ValueAxis axis = null;
		if(metricDistances)
			axis = new TraceDateAxis(trace, (DateEventOutputter)eventOutputter);
		else
			axis = new TraceNumberAxis(trace, eventOutputter);
		//axis.setLabelFont(axisFont);
		axis.setTickLabelFont(axisFont);
		//axis.setAutoRange(true);
		//axis.configure();
		return axis;
	}
	
	
	public double getX(long timestamp, double position) {
		if(metricDistances)
			return timestamp * granularity.getFactor();
		else
			return position;
	}
	
	public int getPosition(RecTrace trace, long timestamp) {
		for(int i = 0; i < trace.getHappenedEvents().size(); i++) {
			if(trace.getHappenedEvents().get(i).getTime() == timestamp)
				return i;
		}
		/*
		for(int i = 0; i < trace.getEvents().size(); i++) {
			
			if(trace.getEvents().get(i).getTime() == timestamp)
				return i;
		}*/
		return -100;
	}	
	
	
	private double getMinDiff(RecTrace trace, TimeGranularity granularity, boolean metricDistances) {
		HappenedEventSet prev = null;
		HappenedEventSet cur = null;
		double minDiff = Long.MAX_VALUE;
		for(int i = 0; i < trace.getHappenedEvents().size(); i++) {
			if(prev == null)
				prev = trace.getHappenedEvents().get(i);
			else {
				cur = trace.getHappenedEvents().get(i);
				double diff = getX(cur.getTime(),i) - getX(prev.getTime(),i-1);
				if(diff > 0 && diff < minDiff)
					minDiff = diff; 
				prev = cur;
			}
		}
		return minDiff;
	}


		
}
