package it.unibo.ai.rec.ui;

import it.unibo.ai.rec.model.FluentGroup;
import it.unibo.ai.rec.model.FluentsModel;
import it.unibo.ai.rec.model.RecTrace;
import it.unibo.ai.rec.model.TrendMetric;

import javax.swing.*;

import org.jfree.chart.ChartPanel;
import org.jfree.chart.plot.XYPlot;

import com.fluxicon.slickerbox.factory.SlickerDecorator;

import java.awt.*;
import java.util.*;

public class FluentChartStandardPanel extends FluentChartContainer {

	
	private static int METRIC_CHART_HEIGHT = 100;
	private static Color metricColor = new Color(47,172,234);
	private JPanel globalContainer;
	private int maximumLabelWidth = 0;
	private TrendMetric metric;
	private JPanel metricPanel;
	
	public FluentChartStandardPanel(FluentChartFactory factory) {
		this(factory, null);
	}
	
	
	public FluentChartStandardPanel(FluentChartFactory factory, TrendMetric metric) {
		super(factory);
		this.metric = metric;
		setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));
		setOpaque(false);
		globalContainer = new JPanel();
		globalContainer.setOpaque(false);
		globalContainer.setLayout(new BoxLayout(globalContainer, BoxLayout.PAGE_AXIS));
		JScrollPane scroller = createSlickerScrollPane();
		JPanel globalWrapper = new JPanel();
		globalWrapper.setLayout(new BoxLayout(globalWrapper, BoxLayout.LINE_AXIS));
		globalWrapper.add(globalContainer);
		globalWrapper.setOpaque(false);
		scroller.getViewport().add(globalContainer);
		add(scroller);
		if(metric != null) {
			metricPanel = new JPanel();
			metricPanel.setOpaque(false);
			metricPanel.setLayout(new BoxLayout(metricPanel, BoxLayout.LINE_AXIS));
			globalContainer.add(metricPanel);
		}
		//add(Box.createVerticalGlue());
		
	}
	
	private JScrollPane createSlickerScrollPane() {
		JScrollPane scrollpane = new JScrollPane();
		scrollpane.setOpaque(false);
		scrollpane.getViewport().setOpaque(false);
		scrollpane.setBorder(BorderFactory.createEmptyBorder());
		//scrollpane.setViewportBorder(BorderFactory.createLineBorder(new Color(
		//		10, 10, 10), 2));
		scrollpane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		scrollpane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
		SlickerDecorator.instance().decorate(scrollpane.getVerticalScrollBar(),
				new Color(0, 0, 0, 0), new Color(140, 140, 140),
				new Color(80, 80, 80));
		scrollpane.getVerticalScrollBar().setOpaque(false);
		SlickerDecorator.instance().decorate(scrollpane.getHorizontalScrollBar(),
				new Color(0, 0, 0, 0), new Color(140, 140, 140),
				new Color(80, 80, 80));
		scrollpane.getHorizontalScrollBar().setOpaque(false);		
		return scrollpane;
	}
	
	
	
	private int getMaximumLabelWidth(Collection<String> labels) {
		String longestLabel = "";
		for(String label: labels)
			if(label.length() > longestLabel.length())
				longestLabel = label;
		return SwingUtilities.computeStringWidth(this.getFontMetrics(this.getFont()), longestLabel);
	}
	
	
	
	
	@Override
	public void update(RecTrace trace, FluentsModel model) {
		globalContainer.removeAll();
		if(metricPanel!=null)
			globalContainer.add(metricPanel);
		
		ArrayList<FluentGroup> groups = new ArrayList<FluentGroup>(model.getFluentGroups().values());
		Collections.sort(groups);		
		
		int height = 0;
		int inset = 0;
		
		globalContainer.removeAll();
		for(int i = 0; i < groups.size(); i++) {
			FluentGroup group = groups.get(i);
			
			ChartPanel p = factory.getFluentChart(trace, group, false, true);
			
			if(i == groups.size() - 1) {
				setShowAxis(p,true);
			}
			else {
				setShowAxis(p,false);
			}
			height += p.getPreferredSize().height;
			
			globalContainer.add(wrapChart(p));
		}
		
		globalContainer.add(Box.createVerticalGlue());
		if(metric != null) {
			metricPanel.removeAll();
			ChartPanel metricChartPanel = factory.getMetricChart(trace, model, metric, METRIC_CHART_HEIGHT);
			XYPlot plot = metricChartPanel.getChart().getXYPlot();
			plot.getDomainAxis().setVisible(false);
			plot.getRenderer().setSeriesPaint(0, metricColor);
			plot.getRenderer().setSeriesStroke(0,new BasicStroke(4));
			metricPanel.add(Box.createHorizontalStrut(inset));
			metricPanel.add(metricChartPanel);
			metricPanel.add(Box.createHorizontalGlue());
			height += metricChartPanel.getPreferredSize().height;
		}
		Dimension d = new Dimension(0,height);
		//this.setPreferredSize(d);
		//this.setMinimumSize(d);
		//this.setMaximumSize(d);
		
	}
	
	private JPanel wrapChart(ChartPanel p) {

		XYPlot plot = p.getChart().getXYPlot();
		String name = plot.getRangeAxis().getLabel();
		plot.getRangeAxis().setLabel("");

		JPanel chartContainer = new JPanel();
		chartContainer.setAlignmentY(Component.TOP_ALIGNMENT);
		chartContainer.setOpaque(false);
		chartContainer.setLayout(new BoxLayout(chartContainer, BoxLayout.PAGE_AXIS));
		
		chartContainer.add(Box.createRigidArea(new Dimension(0,2)));
		
		JLabel nameLabel = new JLabel(name);
		nameLabel.setFont(FluentChartFactory.labelFont);
		nameLabel.setAlignmentX(JLabel.LEFT_ALIGNMENT);
		chartContainer.add(nameLabel);
		
		//add(chartContainer);
		chartContainer.setAlignmentY(Component.TOP_ALIGNMENT);
		//add(Box.createHorizontalGlue());
		p.setAlignmentX(JLabel.LEFT_ALIGNMENT);
		chartContainer.add(p);
		//JPanel detailedChartContainer = new JPanel();
		//detailedChartContainer.setLayout(new BoxLayout(detailedChartContainer, BoxLayout.LINE_AXIS));
		//detailedChartContainer.setAlignmentX(JLabel.LEFT_ALIGNMENT);
		return chartContainer;
	}

	
	private void setShowAxis(ChartPanel chartPanel, boolean showAxis) {
		XYPlot plot = chartPanel.getChart().getXYPlot();
		plot.getDomainAxis().setVisible(showAxis);
		if(showAxis) {
			Dimension d = chartPanel.getPreferredSize();
			int axisHeight = ((HeightEvaluator)plot.getDomainAxis()).getHeight(chartPanel);// d.getFontMetrics(factory.getFluentFont()));
			factory.setDimensions(chartPanel, d.width, d.height+axisHeight);
		}
	}
	
	
	public JPanel getChartPanel() {
		return globalContainer;
	}
}
