package it.unibo.ai.rec.ui;

import java.awt.*;
import java.awt.event.ActionEvent;

import javax.swing.*;

import org.deckfour.uitopia.ui.components.ImageToggleButton;
import org.deckfour.uitopia.ui.util.ImageLoader;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.plot.XYPlot;
import org.jfree.ui.RectangleInsets;

import com.fluxicon.slickerbox.colors.SlickerColors;
import com.fluxicon.slickerbox.components.IconVerticalTabbedPane;
import com.fluxicon.slickerbox.components.SlickerTabbedPane;

import it.unibo.ai.rec.model.*;

public class FluentChartSwitchPanel extends FluentChartContainer implements Comparable<FluentChartSwitchPanel> {

	
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 6787933067618197269L;
	
	private static Font labelFont = new Font("Lucida Grande", Font.PLAIN, 12);
	private String name;
	private JToggleButton viewDetailsButton;
	private ChartPanel summaryChart;
	private ChartPanel detailedChart;
	private JPanel summaryChartContainer;
	private JPanel detailedChartContainer;
	private boolean showAxis;
	private boolean showCommand;
	private FluentGroup currentGroup;
	
	public FluentChartSwitchPanel(FluentChartFactory factory, String name, boolean showAxis, boolean showCommand) {
		super(factory);
		this.showCommand = showCommand;
		setLayout(new BoxLayout(this, BoxLayout.LINE_AXIS));
		this.setOpaque(false);
//		this.setBorder(BorderFactory.createLineBorder(Color.black));
		this.name = name;
		
		if(showCommand) {
		viewDetailsButton = new ImageToggleButton(ImageLoader.load("view_black_20x20.png"), SlickerColors.COLOR_BG_2, Color.white, SlickerColors.COLOR_BG_4, 0);
		viewDetailsButton.setAction(new AbstractAction(){
			@Override
			public void actionPerformed(ActionEvent e) {
				updateView();
			}
		});
		add(viewDetailsButton);
		viewDetailsButton.setAlignmentY(Component.TOP_ALIGNMENT);
		}
		
		JPanel chartContainer = new JPanel();
		chartContainer.setAlignmentY(Component.TOP_ALIGNMENT);
		chartContainer.setOpaque(false);
		chartContainer.setLayout(new BoxLayout(chartContainer, BoxLayout.PAGE_AXIS));
		
		chartContainer.add(Box.createRigidArea(new Dimension(0,2)));
		
		JLabel nameLabel = new JLabel(name);
		nameLabel.setFont(labelFont);
		nameLabel.setFont(FluentChartFactory.labelFont);
		nameLabel.setAlignmentX(JLabel.LEFT_ALIGNMENT);
		chartContainer.add(nameLabel);
		
		add(chartContainer);
		chartContainer.setAlignmentY(Component.TOP_ALIGNMENT);
		add(Box.createHorizontalGlue());
		
		
		summaryChartContainer = new JPanel();
		summaryChartContainer.setLayout(new BoxLayout(summaryChartContainer, BoxLayout.LINE_AXIS));
		summaryChartContainer.setAlignmentX(JLabel.LEFT_ALIGNMENT);
		chartContainer.add(summaryChartContainer);
		
		detailedChartContainer = new JPanel();
		detailedChartContainer.setLayout(new BoxLayout(detailedChartContainer, BoxLayout.LINE_AXIS));
		detailedChartContainer.setAlignmentX(JLabel.LEFT_ALIGNMENT);
		chartContainer.add(detailedChartContainer);
		
		this.showAxis = showAxis;
	}
	
	public void setShowAxis(boolean showAxis) {
		if(this.showAxis != showAxis) {
			this.showAxis = showAxis;
			setupAxes();
		}
	}
	
	private void updateView() {
		if(summaryChart != null)
			summaryChart.setVisible(showCommand && !viewDetailsButton.isSelected());
		if(detailedChart != null)
			detailedChart.setVisible(!showCommand || viewDetailsButton.isSelected());
	}
	
	private void setupAxes() {
		setupAxis(summaryChart);
		setupAxis(detailedChart);
	}
	
	private void setupAxis(ChartPanel chartPanel) {
		XYPlot plot = chartPanel.getChart().getXYPlot();
		plot.getDomainAxis().setVisible(showAxis);
		if(showAxis) {
			Dimension d = chartPanel.getPreferredSize();
			int axisHeight = ((HeightEvaluator)plot.getDomainAxis()).getHeight(chartPanel);// d.getFontMetrics(factory.getFluentFont()));
			factory.setDimensions(chartPanel, d.width, d.height+axisHeight);
		}
	}
	
	private void setupCharts() {
		setupChart(summaryChart);
		setupChart(detailedChart);
	}
	
	private void setupChart(ChartPanel chartPanel) {
		//System.out.println("--------------->"+chartPanel.getFont());

		//chartPanel.getChart().setTitle(name);
		XYPlot plot = chartPanel.getChart().getXYPlot();
		plot.getRangeAxis().setLabel("");
		setupAxis(chartPanel);
	}
	
	@Override
	public void update(RecTrace trace, FluentsModel model) {
		currentGroup = model.getFluentGroups().get(name);
		summaryChartContainer.removeAll();
		detailedChartContainer.removeAll();
		if(currentGroup != null) {
			summaryChart = factory.getSummaryFluentChart(trace, currentGroup, true);
			detailedChart = factory.getDetailedFluentChart(trace, currentGroup, true);
			setupCharts();
			
			summaryChartContainer.add(summaryChart);
			
			detailedChartContainer.add(detailedChart);
			
			updateView();
		}
	}
	
	public FluentGroup getCurrentFluentGroup() {
		return currentGroup;
	}
	
	public int getInset() {
		return (viewDetailsButton==null?0:viewDetailsButton.getWidth());
	}

	@Override
	public int compareTo(FluentChartSwitchPanel p) {
		FluentGroup mine = getCurrentFluentGroup();
		FluentGroup its  = p.getCurrentFluentGroup();
		if(mine == null && its == null)
			return 0;
		else if(mine!=null && its == null)
			return 1;
		else if(mine==null && its != null)
			return -1;
		else
			return mine.compareTo(its);
	}

	@Override
	public JPanel getChartPanel() {
		// TODO Auto-generated method stub
		return null;
	}

}
