package it.unibo.ai.rec.ui;

import it.unibo.ai.rec.model.HappenedEventSet;
import it.unibo.ai.rec.model.RecTrace;

import java.awt.Component;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.swing.SwingUtilities;

import org.jfree.chart.axis.AxisState;
import org.jfree.chart.axis.DateAxis;
import org.jfree.chart.axis.DateTick;
import org.jfree.chart.axis.TickType;
import org.jfree.ui.RectangleEdge;
import org.jfree.ui.TextAnchor;

public class TraceDateAxis extends DateAxis implements HeightEvaluator {

	/**
	 * 
	 */
	private static final long serialVersionUID = -9003477642458529007L;
	RecTrace trace;
	DateEventOutputter eventOutputter;
	
	public TraceDateAxis(RecTrace trace, DateEventOutputter eventOutputter) {
		this.trace = trace;
		this.setVerticalTickLabels(true);
		this.eventOutputter = eventOutputter;
		this.setDateFormatOverride(eventOutputter.getDateFormat());
	}
	
	public List<?> refreshTicks(Graphics2D g2, AxisState state, Rectangle2D dataArea, RectangleEdge edge) {
		List<DateTick> result = new ArrayList<DateTick>();
		for(HappenedEventSet set: trace.getHappenedEvents()) {
			Date date = eventOutputter.getTimestamp(set.getTime());
			result.add(
					new DateTick(
							TickType.MAJOR,
							date,
							eventOutputter.toString(set), 
							TextAnchor.BASELINE_LEFT,
							TextAnchor.BASELINE_LEFT, Math.PI/2)
			);
		}
		return result;
	}

	public int getHeight(Component c) {
		String biggestLabel="";
		for(HappenedEventSet set: trace.getHappenedEvents()) {
			String label = eventOutputter.toString(set);
			if(label.length() > biggestLabel.length())
				biggestLabel = label;
		}
		return SwingUtilities.computeStringWidth(c.getFontMetrics(getTickLabelFont()), biggestLabel+" ");
	}
}
